Index: dart/runtime/bin/eventhandler_linux.cc |
diff --git a/dart/runtime/bin/eventhandler_linux.cc b/dart/runtime/bin/eventhandler_linux.cc |
index 97b305b61e39d05aff432978f25ef07cf846879a..ef0a11dfce65c275e6c48db8251e119e7ed47c44 100644 |
--- a/dart/runtime/bin/eventhandler_linux.cc |
+++ b/dart/runtime/bin/eventhandler_linux.cc |
@@ -6,6 +6,7 @@ |
#if defined(TARGET_OS_LINUX) |
#include "bin/eventhandler.h" |
+#include "bin/eventhandler_linux.h" |
#include <errno.h> // NOLINT |
#include <pthread.h> // NOLINT |
@@ -20,6 +21,7 @@ |
#include "bin/dartutils.h" |
#include "bin/fdutils.h" |
#include "bin/log.h" |
+#include "bin/lockers.h" |
#include "bin/socket.h" |
#include "bin/thread.h" |
#include "platform/utils.h" |
@@ -28,51 +30,52 @@ |
namespace dart { |
namespace bin { |
-static const int kInterruptMessageSize = sizeof(InterruptMessage); |
-static const int kTimerId = -1; |
-static const int kShutdownId = -2; |
- |
-intptr_t SocketData::GetPollEvents() { |
+intptr_t DescriptorInfo::GetPollEvents() { |
// Do not ask for EPOLLERR and EPOLLHUP explicitly as they are |
// triggered anyway. |
intptr_t events = 0; |
- if ((mask_ & (1 << kInEvent)) != 0) { |
+ if ((Mask() & (1 << kInEvent)) != 0) { |
events |= EPOLLIN; |
} |
- if ((mask_ & (1 << kOutEvent)) != 0) { |
+ if ((Mask() & (1 << kOutEvent)) != 0) { |
events |= EPOLLOUT; |
} |
return events; |
} |
-// Unregister the file descriptor for a SocketData structure with epoll. |
-static void RemoveFromEpollInstance(intptr_t epoll_fd_, SocketData* sd) { |
+// Unregister the file descriptor for a DescriptorInfo structure with |
+// epoll. |
+static void RemoveFromEpollInstance(intptr_t epoll_fd_, |
+ DescriptorInfo* di) { |
VOID_NO_RETRY_EXPECTED(epoll_ctl(epoll_fd_, |
EPOLL_CTL_DEL, |
- sd->fd(), |
+ di->fd(), |
NULL)); |
} |
-static void AddToEpollInstance(intptr_t epoll_fd_, SocketData* sd) { |
+static void AddToEpollInstance(intptr_t epoll_fd_, DescriptorInfo* di) { |
struct epoll_event event; |
- event.events = EPOLLRDHUP | sd->GetPollEvents(); |
- if (!sd->IsListeningSocket()) { |
+ event.events = EPOLLRDHUP | di->GetPollEvents(); |
+ if (!di->IsListeningSocket()) { |
event.events |= EPOLLET; |
} |
- event.data.ptr = sd; |
+ event.data.ptr = di; |
int status = NO_RETRY_EXPECTED(epoll_ctl(epoll_fd_, |
EPOLL_CTL_ADD, |
- sd->fd(), |
+ di->fd(), |
&event)); |
if (status == -1) { |
+ // TODO(kustermann): Verify that the dart end is handling this correctly & |
+ // adapt this code to work for multiple listening sockets. |
+ |
// Epoll does not accept the file descriptor. It could be due to |
// already closed file descriptor, or unuspported devices, such |
// as /dev/null. In such case, mark the file descriptor as closed, |
// so dart will handle it accordingly. |
- DartUtils::PostInt32(sd->port(), 1 << kCloseEvent); |
+ DartUtils::PostInt32(di->NextPort(), 1 << kCloseEvent); |
} |
} |
@@ -133,25 +136,26 @@ EventHandlerImplementation::~EventHandlerImplementation() { |
} |
-SocketData* EventHandlerImplementation::GetSocketData(intptr_t fd, |
- bool is_listening) { |
+DescriptorInfo* EventHandlerImplementation::GetDescriptorInfo( |
+ intptr_t fd, bool is_listening) { |
ASSERT(fd >= 0); |
HashMap::Entry* entry = socket_map_.Lookup( |
GetHashmapKeyFromFd(fd), GetHashmapHashFromFd(fd), true); |
ASSERT(entry != NULL); |
- SocketData* sd = reinterpret_cast<SocketData*>(entry->value); |
- if (sd == NULL) { |
+ DescriptorInfo* di = |
+ reinterpret_cast<DescriptorInfo*>(entry->value); |
+ if (di == NULL) { |
// If there is no data in the hash map for this file descriptor a |
// new SocketData for the file descriptor is inserted. |
if (is_listening) { |
- sd = new ListeningSocketData(fd); |
+ di = new DescriptorInfoMultiple(fd); |
} else { |
- sd = new SocketData(fd); |
+ di = new DescriptorInfoSingle(fd); |
} |
- entry->value = sd; |
+ entry->value = di; |
} |
- ASSERT(fd == sd->fd()); |
- return sd; |
+ ASSERT(fd == di->fd()); |
+ return di; |
} |
@@ -197,38 +201,62 @@ void EventHandlerImplementation::HandleInterruptFd() { |
} else if (msg[i].id == kShutdownId) { |
shutdown_ = true; |
} else { |
- SocketData* sd = GetSocketData( |
+ DescriptorInfo* di = GetDescriptorInfo( |
msg[i].id, (msg[i].data & (1 << kListeningSocket)) != 0); |
if (IS_COMMAND(msg[i].data, kShutdownReadCommand)) { |
- ASSERT(!sd->IsListeningSocket()); |
+ ASSERT(!di->IsListeningSocket()); |
// Close the socket for reading. |
- VOID_NO_RETRY_EXPECTED(shutdown(sd->fd(), SHUT_RD)); |
+ VOID_NO_RETRY_EXPECTED(shutdown(di->fd(), SHUT_RD)); |
} else if (IS_COMMAND(msg[i].data, kShutdownWriteCommand)) { |
- ASSERT(!sd->IsListeningSocket()); |
+ ASSERT(!di->IsListeningSocket()); |
// Close the socket for writing. |
- VOID_NO_RETRY_EXPECTED(shutdown(sd->fd(), SHUT_WR)); |
+ VOID_NO_RETRY_EXPECTED(shutdown(di->fd(), SHUT_WR)); |
} else if (IS_COMMAND(msg[i].data, kCloseCommand)) { |
// Close the socket and free system resources and move on to next |
// message. |
- if (sd->RemovePort(msg[i].dart_port)) { |
- RemoveFromEpollInstance(epoll_fd_, sd); |
- intptr_t fd = sd->fd(); |
- sd->Close(); |
- socket_map_.Remove(GetHashmapKeyFromFd(fd), GetHashmapHashFromFd(fd)); |
- delete sd; |
+ bool no_more_listeners = di->RemovePort(msg[i].dart_port); |
+ if (no_more_listeners) { |
+ RemoveFromEpollInstance(epoll_fd_, di); |
+ } |
+ |
+ intptr_t fd = di->fd(); |
+ if (di->IsListeningSocket()) { |
+ // We only close the socket file descriptor from the operating |
+ // system if there are no other dart socket objects which |
+ // are listening on the same (address, port) combination. |
+ { |
+ MutexLocker ml(globalTcpListeningSocketRegistry.mutex()); |
+ if (globalTcpListeningSocketRegistry.CloseSafe(fd)) { |
+ socket_map_.Remove( |
+ GetHashmapKeyFromFd(fd), GetHashmapHashFromFd(fd)); |
+ di->Close(); |
+ delete di; |
+ } |
+ } |
+ } else { |
+ ASSERT(no_more_listeners); |
+ socket_map_.Remove( |
+ GetHashmapKeyFromFd(fd), GetHashmapHashFromFd(fd)); |
+ di->Close(); |
+ delete di; |
} |
DartUtils::PostInt32(msg[i].dart_port, 1 << kDestroyedEvent); |
} else if (IS_COMMAND(msg[i].data, kReturnTokenCommand)) { |
int count = TOKEN_COUNT(msg[i].data); |
- if (sd->ReturnToken(msg[i].dart_port, count)) { |
- AddToEpollInstance(epoll_fd_, sd); |
+ if (di->ReturnTokens(msg[i].dart_port, count)) { |
+ AddToEpollInstance(epoll_fd_, di); |
} |
} else { |
ASSERT_NO_COMMAND(msg[i].data); |
- // Setup events to wait for. |
- if (sd->AddPort(msg[i].dart_port)) { |
- sd->SetMask(msg[i].data); |
- AddToEpollInstance(epoll_fd_, sd); |
+ bool had_listeners = di->HasNextPort(); |
+ di->SetPortAndMask(msg[i].dart_port, msg[i].data & EVENT_MASK); |
+ bool has_listeners = di->HasNextPort(); |
+ |
+ // Add/Remove from epoll set depending on previous and current state. |
+ if (!had_listeners && has_listeners) { |
+ AddToEpollInstance(epoll_fd_, di); |
+ } else if (had_listeners && !has_listeners) { |
+ RemoveFromEpollInstance(epoll_fd_, di); |
} |
} |
} |
@@ -256,9 +284,9 @@ static void PrintEventMask(intptr_t fd, intptr_t events) { |
#endif |
intptr_t EventHandlerImplementation::GetPollEvents(intptr_t events, |
- SocketData* sd) { |
+ DescriptorInfo* di) { |
#ifdef DEBUG_POLL |
- PrintEventMask(sd->fd(), events); |
+ PrintEventMask(di->fd(), events); |
#endif |
if (events & EPOLLERR) { |
// Return error only if EPOLLIN is present. |
@@ -287,15 +315,16 @@ void EventHandlerImplementation::HandleEvents(struct epoll_event* events, |
timeout_queue_.RemoveCurrent(); |
} |
} else { |
- SocketData* sd = reinterpret_cast<SocketData*>(events[i].data.ptr); |
- intptr_t event_mask = GetPollEvents(events[i].events, sd); |
+ DescriptorInfo* di = |
+ reinterpret_cast<DescriptorInfo*>(events[i].data.ptr); |
+ intptr_t event_mask = GetPollEvents(events[i].events, di); |
if (event_mask != 0) { |
- Dart_Port port = sd->port(); |
- if (sd->TakeToken()) { |
+ Dart_Port port = di->NextPort(); |
+ ASSERT(port != 0); |
+ if (di->TakeToken()) { |
// Took last token, remove from epoll. |
- RemoveFromEpollInstance(epoll_fd_, sd); |
+ RemoveFromEpollInstance(epoll_fd_, di); |
} |
- ASSERT(port != 0); |
DartUtils::PostInt32(port, event_mask); |
} |
} |
@@ -333,7 +362,7 @@ void EventHandlerImplementation::Poll(uword args) { |
void EventHandlerImplementation::Start(EventHandler* handler) { |
int result = Thread::Start(&EventHandlerImplementation::Poll, |
- reinterpret_cast<uword>(handler)); |
+ reinterpret_cast<uword>(handler)); |
if (result != 0) { |
FATAL1("Failed to start event handler thread %d", result); |
} |
@@ -133,25 +136,26 @@ EventHandlerImplementation::~EventHandlerImplementation() { |
// If there is no data in the hash map for this file descriptor a |
- // new SocketData for the file descriptor is inserted. |
+ // new DescriptorInfo for the file descriptor is inserted. |
if (is_listening) { |